import os
from pathlib import Path
import click

from .config import Config
from .core import (
    notion_client,
    notion_blog_database,
    page_to_markdown,
    page_path,
    published_pages,
    draft_pages,
)
from .version import __version__

"""
`cli` Module only has thin wrappers around Notoma Python API
that invokes the API with provided configuration.

Run `notoma` for the help article on how to use the CLI.
"""


# The CLI runner
#
@click.group(
    help="""Build your staticg gen blog with Notion.
    Notoma converts Notion database of blog posts into a directory of .md files."""
)
def runner() -> None:
    pass


@runner.command(help="Print Notoma version.")
def version():
    click.echo(f"Notoma {__version__}")


@runner.command(help="Convert Notion Blog to Markdown files.")
@click.option("--from", "-f", "notion_url", help="Notion blog URL")
@click.option(
    "--dest",
    "-d",
    required=True,
    default="posts",
    type=click.Path(exists=True),
    help="Directory to put posts into.",
)
@click.option(
    "--drafts",
    default=None,
    type=click.Path(exists=True),
    help="Directory for draft posts. Drafts won't be imported if this is left blank.",
)
@click.option("--token_v2", "-t", help="Notion auth token from the cookie.")
@click.option("--verbose", is_flag=True, default=False, help="Verbose output.")
@click.option("--debug", is_flag=True, default=False, help="Enable debug output.")
def convert(
    dest: str,
    drafts: str,
    debug: bool = False,
    verbose: bool = False,
    token_v2: str = None,
    notion_url: str = None,
) -> None:
    config = Config(token_v2=token_v2, blog_url=notion_url)
    __validate_config(config)

    dest = Path(dest).absolute()
    client = notion_client(config.token_v2)
    blog = notion_blog_database(client, config.blog_url)

    if verbose:
        click.echo(f"Processing articles from Notion: {blog.parent.title}")

    __convert_pages(published_pages(blog), dest, verbose)

    if drafts:
        __convert_pages(draft_pages(blog), Path(drafts).absolute(), verbose)
        draft_pages(blog)


@runner.command()
def watch() -> None:
    """
    Watch for updates in the Notion Blog and update the Markdown posts
    """
    raise NotImplementedError("Watcher is not implemented yet.")


@runner.command()
def new() -> None:
    """
    Create a new Notion Blog
    """
    raise NotImplementedError("Creating a new blog is not implemented yet.")


def __convert_pages(pages: list, dest_dir: Path, verbose: bool = False) -> None:
    "Convert a bunch of pages with a nice progress bar."

    if verbose:
        click.echo(f"{len(pages)} pages to process.")

    with click.progressbar(pages) as bar:
        for page in bar:
            page_path(page, dest_dir=dest_dir).write_text(page_to_markdown(page))

    if verbose:
        click.echo(f"Processed {len(pages)} pages.")


def __validate_config(config: Config) -> None:
    """
    Validates the provided options and prints errors to stdout,
    then aborts if there are any errors.
    """
    errors = list()
    if config.token_v2 is None:
        errors.append(
            "Error: Authentication token `token_v2` is not provided. Try --token option."
        )

    if config.blog_url is None:
        errors.append(
            "Error: Notion Blog Database URL not provided. Try --from option."
        )

    if len(errors) > 0:
        for e in errors:
            click.echo(e)
        raise click.Abort()



import yaml
from pathlib import Path
from typing import Union, List

from notion.client import NotionClient
from notion.block import PageBlock
from notion.collection import Collection, NotionDate

from .config import Config
from .templates import template


def notion_client(token_v2: str) -> NotionClient:
    config = Config(token_v2=token_v2)
    client = NotionClient(token_v2=config.token_v2)
    return client


def notion_blog_database(client: NotionClient, db_url: str) -> Collection:
    """
    Returns a Notion database, wraped into a `notion.Collection` for
    easy access to it's rows.
    """
    return client.get_collection_view(db_url).collection


def all_pages(blog: Collection) -> List[PageBlock]:
    "Returns all Notion PageBlocks"
    return blog.get_rows()


def published_pages(blog: Collection) -> List[PageBlock]:
    "Returns the list of published pages."
    # FIXME: This needs to be a filtered query instead.
    return [post for post in blog.get_rows() if post.get_all_properties()["published"]]


def draft_pages(blog: Collection) -> List[PageBlock]:
    "Returns the list of draft pages."
    # FIXME: This needs to be a filtered query instead.
    return [
        post for post in blog.get_rows() if not post.get_all_properties()["published"]
    ]


def page_to_markdown(page: PageBlock) -> str:
    "Translates a Notion Page (`PageBlock`) into a Markdown string."
    return template("post", debug=True).render(
        page=page, front_matter=front_matter(page)
    )


def page_path(page: PageBlock, dest_dir: Path = Path(".")) -> Path:
    "Build a .md file path in `dest_dir` based on a Notion page metadata."
    fname = "-".join(page.title.lower().replace(".", "").split(" ")) + ".md"
    return dest_dir / fname


def front_matter(page: PageBlock, config: Config = Config()) -> str:
    "Builds and returns a page front matter in a yaml-like format."
    all_props = page.get_all_properties()
    if "layout" not in all_props:
        all_props["layout"] = config.default_layout
    renderables = {k: v for k, v in all_props.items() if v != ""}
    return __sanitize_front_matter(renderables)


def __sanitize_front_matter(items: dict) -> dict:
    "Sanitizes and returns front matter items."
    for k, v in items.items():
        if type(v) not in [str, list]:
            if isinstance(v, NotionDate):
                items[k] = v.start
            if isinstance(v, bool):
                items[k] = str(v).lower()
    return items



from typing import Union
from pathlib import Path
from jinja2 import (
    Environment,
    Template,
    PackageLoader,
    select_autoescape,
    contextfilter,
)
from jinja2.runtime import Context

import notion.block as block
import re

"""
Provides templates that are used in Notion -> Markdown conversion.
They're currently not utilized, as the render engine actually uses
`notion.markdown.notion_to_markdown` instead
"""


def template(name: Union[str, Path], debug: bool = False) -> Template:
    """
    Loads the template file `templates/{name}.md.j2`.
    """

    filters = dict(
        template_name=__template_name,
        render_block=__render_block,
        numbered_list_index=__numbered_list_index,
        notion_url=__notion_url,
    )

    env = Environment(
        loader=PackageLoader("notoma", "templates"),
        autoescape=select_autoescape(["md"]),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    for fl, func in filters.items():
        env.filters[fl] = func
    env.globals["debug"] = debug

    return env.get_template(f"{name}.md.j2")


def __template_name(block: block.Block) -> str:
    """
    Finds a Jinja template name for the provided Notion `Block`,
    checks that the template exists, and returns the template name as `str`.
    """
    template_name = f"blocks/_{__block_type(block)}.md.j2"
    template_path = Path(__file__).parent / f"templates/{template_name}"

    if template_path.exists():
        return template_name
    else:
        return None


@contextfilter
def __render_block(ctx: Context, block: block.Block) -> str:
    """
    Jinja filter that wraps the block in it's markdown equivalent if possible,
    or returns it's title.
    """
    if ctx["debug"]:
        print(f"Unsupported block type: {__block_type(block)} in {ctx['page'].title}.")
    return str(block.title)


@contextfilter
def __numbered_list_index(ctx: Context, this: block.Block) -> str:
    index = 1
    for item in ctx["page"].children:
        if isinstance(item, block.NumberedListBlock):
            if item.id == this.id:
                return index
            else:
                index += 1
        else:
            index = 1
    raise ValueError(
        f"Expected the page {ctx['page'].id} to include the provided block {this.id}"
    )


def __block_type(block: block.Block) -> str:
    "Returns snake_cased block name."
    return __snake_case(block.__class__.__name__)[:-6]


def __snake_case(input_string: str) -> str:
    "Returns a new string containing the input string, in camel_case."
    output = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", input_string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", output).lower()


def __notion_url(page: block.PageBlock) -> str:
    "Returns the page's Notion URL."
    return page.get_browseable_url()



import os
from setuptools import setup, find_packages
from notoma.version import __version__ as notoma_version

ROOT = os.path.dirname(os.path.abspath(__file__))


def get_requirements(env):
    """
    Takes requirements from requirements.txt.
    The detault (produciton) environment is "base"
    """
    with open(f"requirements-{env}.txt") as fp:
        reqs = list()
        for lib in fp.read().split("\n"):
            if not lib.startswith("-") or lib.startswith("#"):
                reqs.append(lib.strip())
        return reqs


setup(
    name="notoma",
    version=notoma_version,
    author="Nate Gadzhibalaev",
    author_email="nate@respawn.io",
    url="https://github.com/xnutsive/notoma/",
    description="Write your blog articles in Notion. Notoma converts your Notion database pages to Markdown files.",
    long_description=open(os.path.join(ROOT, "README.md")).read(),
    long_description_content_type="text/markdown",
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=get_requirements("base"),
    extras_require={"dev": get_requirements("dev")},
    license="Apache Software License 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Typing :: Typed",
    ],
    project_urls={
        "Documentation": "https://xnutsive.github.io/notoma/",
        "Source Code": "https://github.com/xnutsive/notoma/",
    },
    entry_points={
        "console_scripts": [
            "notoma-dev = notoma.dev:cli",
            "notoma = notoma.cli:runner",
        ]
    },
    include_package_data=True,
    packages=find_packages(),
)



import os
from dotenv import load_dotenv, find_dotenv


CONF_MAP = dict(
    token_v2="NOTOMA_NOTION_TOKEN_V2",
    blog_url="NOTOMA_NOTION_BLOG_URL",
    default_layout="NOTOMA_DEFAULT_LAYOUT",
)


class Config:
    """
    Wraps Notoma's settings in an object.
    Settings are automatically loaded from ENV (and `.env` file), and you
    can override them with `kwargs`.

        - `token_v2`: str, Notion authentication token. Environment variable
            `NOTOMA_NOTION_TOKEN_V2`.
        - `blog_url`: str, Notion Blog URL. `NOTOMA_NOTION_BLOG_URL`.
    """

    def __init__(self, **kwargs):
        """
        Loads config from a `.env` file or system environment.

        You can provide any kwargs you want and they would override
         environment config values.
        """
        load_dotenv(find_dotenv())

        self.__config = {k: os.environ.get(v) for k, v in CONF_MAP.items()}

        for key, value in kwargs.items():
            if value is not None:
                self.__config[key] = value

    @property
    def token_v2(self) -> str:
        return self.__config["token_v2"]

    @property
    def blog_url(self) -> str:
        return self.__config["blog_url"]

    @property
    def default_layout(self) -> str:
        return self.__config["default_layout"]

    def __getitem__(self, key):
        return self.__config[key]

    def __repr__(self):
        return "\n".join(f"{k}: {v}" for k, v in self.__config.items())
